'''
------------------------------------------------------------------------------
Name:         CategoryManager.py
Purpose:      CategoryManager provides a treeish category structure to
              assoiate objects to.  The tree is made up of folderish Category
              node objects, which provide recursive functions.
Author:       Jeremy Stark <jeremy@deximer.com>
              Joseph Zicarelli <jdz@deximer.com>
Copyright:    (c) 2005 by Deximer
Licence:      GNU General Public Licence (GPL) Version 2 or later
------------------------------------------------------------------------------
'''

from sets import Set

from Products.CMFCore.utils import UniqueObject
from Products.CMFCore.utils import getToolByName
from Products.CMFCore import CMFCorePermissions
from AccessControl import ClassSecurityInfo

from Products.Archetypes.public import *
from Products.Archetypes import listTypes

from Products.PortalTaxonomy.interfaces.ICategorizable import ICategorizable

class CategoryManager(BaseContent, UniqueObject):
    security = ClassSecurityInfo()
    portal_type = meta_type = archetype_name = 'CategoryManager' 
    schema=BaseFolderSchema + Schema((),)

    def __init__(self):
        BaseContent.__init__(self, 'portal_categories')
        self.setTitle('Category Tool')

    def recurseTypes(self):
        ''' Stopper function: this is the bottom.  Simply return.
            NOTE: Not needed?  CategoryManager no longer top of tree
        '''
        return []

    def getTemplate(self):
        ''' Stopper function for Category.getTemplate()
        '''
        return ['category_listing',]

    def getSubCategories(self, type=None):
        ''' Lists all Category objects directly within this Category
            NOTE: This should do a catalog query for isRootCategory Categories
        '''
        cats = getToolByName(self, 'portal_url').getPortalObject().objectValues(['Category',])
        if type is None:
            return cats
        result = []
        type_name = type.getTypeInfo().Metatype() + ':' + type.getName()
        for cat in cats:
            if cat.testTypes(type_name):
                result.append(cat)
        return result

    def getVocab(self, instance, depth=None):
        ''' Wrapper for getVocabTree(), returns DisplayList for vocabs
            NOTE: This should do a catalog query for isRootCategory Categories
        '''
        cats = getToolByName(self, 'portal_url').getPortalObject().objectValues(['Category',])
        root = Category('TempRootNode').__of__(self)

        # Note: Appending _temp to end of cat ids so they do not conflict
        # with any cats in the portal root.  This is ok because we create
        # an association with the object using the cat UIDs, not ids.
        for cat in cats:
            root._setObject(cat.id + '_temp', cat)
        return DisplayList(root.getVocabTree(instance, depth=depth))

    def testTypes(self, type):
        """
        types -- list from the types restriction field
        type -- the type of an object we're testing agenst

        return True if:
         * types is empty (or null)
         * type is in types -- a match!
        """
        types = self.getTypes()
        if not types or type in types:
            return True
        return False

    def getRootCategories(self):
        ''' Return published Categories that are also rootNodes
        '''
        pass

    factory_type_information={
        'allowed_content_types':('Category', ),
        'content_icon':'categorymanager.gif',
        'immediate_view':'folder_contents',
        'global_allow':0,
        'filter_content_types':1
        }

class Category(BaseFolder):
    security = ClassSecurityInfo()
    portal_type = meta_type = archetype_name = 'Category' 
    right_slots=[] # disable if you want right slots when viewing categories
    schema=BaseFolderSchema  + Schema((
        LinesField(
            'Types'
          , widget = MultiSelectionWidget(
                size = 10
          , description = "Select the content types allowed in this Category.<br>\n <i>Note: selecting nothing is a special case - if you select no types, this category will inherit types from above categories.</i>"
              , label = "Type Restriction"
            )
          , vocabulary = "getTypesVocab"
          , mutator = "setTypes"
          , enforceVocabulary = True
        ),
        LinesField(
            'render_template'
          , widget = SelectionWidget(label = 'Render Template' )
          , vocabulary = 'getRenderTemplateVocab'
          , enforceVocabulary = True
        )
    ),)

    def setTypes(self, value, parent=None):
        ''' Recursive mutator to update category types.
            This is needed because of dynamic vocabularies
        '''
        value = [x for x in value if x]  # remove None values
        self.Types = value  # store the new value

        if value:
            parent = value
        if not parent:
            return

        for cat in self.getSubCategories():
            cat.setTypes(Set(parent).intersection(cat.getTypes()), parent)

    def recurseTypes(self):
        ''' Two cases:
             -types is empty: ask parent
             -types is not empty: return types
        '''
        myTypes = self.getTypes()
        if myTypes:
            return myTypes
        else:
            if hasattr(self.aq_parent, 'recurseTypes'):
                return self.aq_parent.recurseTypes()
            else:
                return []

    def getTypesVocab(self):
        ''' Wrapper for recursive recurseTypes(), compares to AvailableTypes
        '''
        parentTypes = []
        if hasattr(self.aq_parent, 'recurseTypes'):
            parentTypes = self.aq_parent.recurseTypes()
        if parentTypes:
            return parentTypes
        else:
            return self.getAvailableTypes()

    def getAvailableTypes(self):
        ''' Return types that have a CategoryField in their schema
        '''
        result = []
        portalTypes = getToolByName(self, 'portal_types').listTypeInfo()
        pt = [t.Metatype() for t in portalTypes]
        types = [t for t in listTypes() if t['klass'].meta_type in pt]
        for type in types:
            schema = type['klass'].schema
            for field in schema.values():
                if field.getType()[-13:] == 'CategoryField':
                    result.append(type['klass'].meta_type + ':' + field.getName())
        return DisplayList([(x, x) for x in result])

    def getSubCategories(self, type=None):
        ''' Lists all Category objects directly within this Category
        '''
        cats = self.listFolderContents()
        if type is None:
            return cats
        result = []
        type_name = type.getTypeInfo().Metatype()
        for cat in cats:
            if cat.testTypes(type_name):
                result.append(cat)
        return result

    def getContent(self, review_state = 'published'):
        ''' Returns published site content assoiated with this Category
        '''
        catalog = getToolByName(self, 'portal_catalog')
        return catalog.searchResults(
            getCategories = self.UID()
          , review_state = review_state
          )

    def getVocabTree(self, instance, depth=None, path=''):
        ''' Recursevly lists Categories up to `depth' Categories deep.
            Output tailored for creation of a DisplayList.
        '''
        cats = []
        type_field = instance.meta_type + ':' + instance.getName()
        if (depth is None) or (depth > 0):
            if depth is not None:
                depth -= 1
            for cat in self.getSubCategories():
                if cat.testTypes(type_field):
                    guid = cat.UID()
                    title = cat.title
                    localPath = path + title
                    cats.append([guid, localPath])
                    cats += list(cat.getVocabTree(instance, depth, localPath + '/'))
        return cats

    def getVocab(self, instance, depth=None):
        ''' Wrapper for getVocabTree(), returns DisplayList for vocabs
            NOTE: This may need to look at instance.metatype
        '''
        return DisplayList(self.getVocabTree(instance, depth=depth))

    def testTypes(self, type):
        """
        types -- list from the types restriction field
        type -- the type of an object we're testing agenst

        return True if:
         * types is empty (or null)
         * type is in types -- a match!
        """
        types = self.getTypes()
        if not types or type in types:
            return True
        return False

    def isRootCategory(self):
        ''' Returns True if this is a root node
        '''
        return not hasattr(self.aq_parent, 'isRootCategory')

    def getRenderTemplateVocab(self):
        ''' Return a list of available page templates to render this category
            Should use portal_actions getActionsFor type call so that TTW
            actions are included in the available templates
            NOTE: This should actually relly on Plone 2.1 template system
        '''
        result = [('', 'Parent')]
        for action in self.actions:
            if action['category'] == 'public_view':
                result.append((action['action'], action['name']))
        return DisplayList(result)

    def index_html(self):
        ''' Renders the public view of the category, usualy a recursive template
        '''
        return self.aq_acquire(self.getTemplate()[0])(self)

    def getTemplate(self):
        ''' Return a template for rendering this category
        '''
        if not self.render_template and hasattr(self.aq_parent, 'getTemplate'):
            return self.aq_parent.getTemplate()
        else:
            return ['general_listing',]
        return self.render_template

    actions=(
    { 'id': 'category_listing'
    , 'name': 'Category Listing'
    , 'action': 'category_listing'
    , 'visible': 0
    , 'permissions': (CMFCorePermissions.View,)
    , 'category': 'object'
    },
    { 'id': 'category_recursive'
    , 'name': 'Recursive Category Listing'
    , 'action': 'category_recursive'
    , 'visible': 0
    , 'permissions': (CMFCorePermissions.View,)
    , 'category': 'object'
    },
    { 'id': 'category_utilities'
    , 'name': 'Utilities'
    , 'action': 'category_utilities'
    , 'visible': 1
    , 'permissions': (CMFCorePermissions.ManagePortal,)
    , 'category': 'object'
    },
    { 'id': 'general_listing'
    , 'name': 'General Listing'
    , 'action': 'directory'
    , 'visible': 0
    , 'permissions': (CMFCorePermissions.View,)
    , 'category': 'public_view'
    },
    )

    factory_type_information={
        'allowed_content_types':('Category', ) ,
        'content_icon':'category.gif',
        'immediate_view':'folder_contents',
        'global_allow':1,
        'filter_content_types':1
        }

registerType(CategoryManager)
registerType(Category)
